from django.conf.urls import url

from .views import GuestListView, GuestListJsonView

urlpatterns = [
    url(r'^json/(?P<guestlist_id>[0-9\+]+)/$', GuestListJsonView.as_view(), name='guestListJSON'),
    url(r'^json/(?P<guestlist_id>[0-9\+]+)/(?P<event_id>[0-9\+]+)/$', GuestListJsonView.as_view(), name='guestListJSON'),
    url(r'^(?P<guestlist_id>[0-9\+]+)/$', GuestListView.as_view(), name='viewGuestList'),
    url(r'^(?P<guestlist_id>[0-9\+]+)/(?P<event_id>[0-9\+]+)/$', GuestListView.as_view(), name='viewGuestList'),
]



from django.utils.translation import ugettext_lazy as _


GUESTLIST_ADMISSION_CHOICES = [
    ('Always',_('Always added to guest list')),
    ('EventOnly',_('Add if the person is a staff member for this event')),
    ('Day',_('Add if the person is a staff member on that day')),
    ('Week',_('Add if the person is a staff member in that week')),
    ('Month',_('Add if the person is a staff member in that month')),
    ('Year',_('Add if the person is a staff member in that year')),
]


GUESTLIST_SORT_CHOICES = [
    ('Last',_('Last name (default)')),
    ('First',_('First name')),
    ('Comp',_('Admission type (e.g. staff, registrant, other')),
]



from django.utils.translation import ugettext_lazy as _
from django.views.generic import DetailView
from django.shortcuts import get_object_or_404
from django.http import Http404, JsonResponse

from braces.views import PermissionRequiredMixin

from .models import GuestList, Event


class GuestListView(PermissionRequiredMixin,DetailView):
    '''
    This view is used to access the guest list for a given series or event
    '''
    template_name = 'guestlist/guestlist.html'
    permission_required = 'guestlist.view_guestlist'

    def get_object(self, queryset=None):
        ''' Get the guest list from the URL '''
        return get_object_or_404(
            GuestList.objects.filter(id=self.kwargs.get('guestlist_id')))

    def get_context_data(self,**kwargs):
        ''' Add the list of names for the given guest list '''
        event = Event.objects.filter(id=self.kwargs.get('event_id')).first()
        if self.kwargs.get('event_id') and not self.object.appliesToEvent(event):
            raise Http404(_('Invalid event.'))

        # Use the most current event if nothing has been specified.
        if not event:
            event = self.object.currentEvent

        context = {
            'guestList': self.object,
            'event': event,
            'names': self.object.getListForEvent(event),
        }
        context.update(kwargs)
        return super(GuestListView,self).get_context_data(**context)


class GuestListJsonView(GuestListView):
    '''
    This view provides the guest list in JSON format.
    '''
    template_name = None

    def render_to_response(self, context, **response_kwargs):
        json_context = {
            'names': context.get('names'),
            'guestlist_id': getattr(context.get('guestList'),'id'),
            'event_id': getattr(context.get('event'),'id'),            
        }

        return JsonResponse(json_context)



from django.core.exceptions import ValidationError
from django.db import models
from django.db.models import Q, Value, Case, When
from django.db.models.functions import Concat
from django.utils.translation import ugettext_lazy as _
from django.utils import timezone

from cms.models.pluginmodel import CMSPlugin
from intervaltree import IntervalTree
from datetime import timedelta
from dateutil.relativedelta import relativedelta

from danceschool.core.models import PublicEventCategory, SeriesCategory, EventStaffCategory, EventSession, Event, StaffMember, Registration
from danceschool.core.utils.timezone import ensure_localtime
from .constants import GUESTLIST_ADMISSION_CHOICES, GUESTLIST_SORT_CHOICES


class GuestList(models.Model):

    name = models.CharField(_('Name'),max_length=200,unique=True)
    sortOrder = models.CharField(_('Sort order'),choices=GUESTLIST_SORT_CHOICES,default='Last',max_length=5)

    # Rules for which events a guest list applies to
    seriesCategories = models.ManyToManyField(SeriesCategory,verbose_name=_('Series categories'),blank=True)
    eventCategories = models.ManyToManyField(PublicEventCategory,verbose_name=_('Public event categories'),blank=True)
    eventSessions = models.ManyToManyField(EventSession,verbose_name=_('Event sessions'),blank=True)
    individualEvents = models.ManyToManyField(Event,verbose_name=_('Individual events'),blank=True,related_name='specifiedGuestLists')

    includeStaff = models.BooleanField(_('Include all scheduled event staff'),default=True,blank=True)
    includeRegistrants = models.BooleanField(_('Include event registrants'),default=True,blank=True)

    @property
    def recentEvents(self):
        '''
        Get the set of recent and upcoming events to which this list applies.
        '''
        return Event.objects.filter(
            Q(pk__in=self.individualEvents.values_list('pk',flat=True)) |
            Q(session__in=self.eventSessions.all()) |
            Q(publicevent__category__in=self.eventCategories.all()) |
            Q(series__category__in=self.seriesCategories.all())
        ).filter(
            Q(startTime__lte=timezone.now() + timedelta(days=60)) &
            Q(endTime__gte=timezone.now() - timedelta(days=60))
        )

    @property
    def currentEvent(self):
        '''
        Return the first event that hasn't ended yet, or if there are no
        future events, the last one to end.
        '''
        currentEvent =  self.recentEvents.filter(endTime__gte=timezone.now()).order_by('startTime').first()
        if not currentEvent:
            currentEvent = self.recentEvents.filter(
                endTime__lte=timezone.now()
            ).order_by('-endTime').first()
        return currentEvent

    def appliesToEvent(self, event):
        ''' Check whether this guest list is applicable to an event. '''
        return (
            event in self.individualEvents.all() or
            event.session in self.eventSessions.all() or
            event.category in self.seriesCategories.all() or
            event.category in self.eventCategories.all()
        ) 

    def getDayStart(self, dateTime):
        ''' Ensure local time and get the beginning of the day '''
        return ensure_localtime(dateTime).replace(hour=0,minute=0,second=0,microsecond=0)

    def getComponentFilters(self,component,event=None,dateTime=None):
        '''
        Get a parsimonious set of intervals and the associated Q() objects
        based on the occurrences of a specified event, and the rule that
        implicitly defines the start and end of each interval.
        '''

        # Limit to the staff member or staff category specified by the rule.
        if component.staffMember:
            filters = Q(pk=component.staffMember.pk)
        else:
            filters = Q(eventstaffmember__category=component.staffCategory)

        # Handle 'Always' and 'EventOnly' rules first, because they do not require an analysis of intervals.
        if component.admissionRule == 'EventOnly' and event:
            # Skip the analysis of intervals and include only those who are staffed for the event.
            return Q(filters & Q(eventstaffmember__event=event))
        elif component.admissionRule in ['Always','EventOnly']:
            # If 'Always' or no event is specified, include all associated staff
            return Q(filters)

        # Start with the event occurrence intervals, or with the specified time.
        if event:
            intervals = [(x.startTime,x.endTime) for x in event.eventoccurrence_set.all()]
        elif dateTime:
            intervals = [(dateTime,dateTime)]
        else:
            raise ValueError(_('Must provide either an event or a datetime to get interval queries.'))

        if component.admissionRule == 'Day':
            # The complete days of each event occurrence
            intervals = [
                (self.getDayStart(x[0]),self.getDayStart(x[1]) + timedelta(days=1)) for x in intervals
            ]
        elif component.admissionRule == 'Week':
            # The complete weeks of each event occurrence
            intervals = [
                (
                    self.getDayStart(x[0]) - timedelta(days=x[0].weekday()),
                    self.getDayStart(x[1]) - timedelta(days=x[1].weekday() - 7)
                ) for x in intervals
            ]
        elif component.admissionRule == 'Month':
            # The complete month of each event occurrence
            intervals = [
                (
                    self.getDayStart(x[0]).replace(day=1),
                    self.getDayStart(x[1]).replace(day=1) + relativedelta(months=1)
                ) for x in intervals
            ]
        elif component.admissionRule == 'Year':
            # The complete years of each event occurrence
            intervals = [
                (
                    self.getDayStart(x[0]).replace(month=1,day=1),
                    self.getDayStart(x[1]).replace(year=x[1].year + 1,month=1,day=1)
                ) for x in intervals
            ]
        else:
            # This is a failsafe that will always evaluate as False.
            return Q(pk__isnull=True)

        # Use intervaltree to create the most parsimonious set of intervals for this interval
        # and then filter on those intervals
        intervals = [sorted(x) for x in intervals]
        tree = IntervalTree.from_tuples(intervals)
        tree.merge_overlaps()

        # Since we are OR appending, start with something that is always False.
        intervalFilters = Q(pk__isnull=True)

        for item in tree.items():
            intervalFilters = intervalFilters | Q(
                Q(eventstaffmember__event__eventoccurrence__endTime__gte=item[0]) &
                Q(eventstaffmember__event__eventoccurrence__startTime__lte=item[1])
            )

        return Q(filters & intervalFilters)

    def getListForEvent(self, event=None):
        ''' Get the list of names associated with a particular event. '''
        names = list(self.guestlistname_set.annotate(
            guestType=Value('Manually Added',output_field=models.CharField())
        ).values('firstName','lastName','guestType'))

        # Component-by-component, OR append filters to an initial filter that always
        # evaluates to False.
        components = self.guestlistcomponent_set.all()
        filters = Q(pk__isnull=True)

        # Add prior staff based on the component rule.
        for component in components:
            if event and self.appliesToEvent(event):
                filters = filters | self.getComponentFilters(component,event=event)
            else:
                filters = filters | self.getComponentFilters(component,dateTime=timezone.now())

        # Add all event staff if that box is checked (no need for separate components)
        if self.includeStaff and event and self.appliesToEvent(event):
            filters = filters | Q(eventstaffmember__event=event)

        # Execute the constructed query and add the names of staff
        names += list(StaffMember.objects.filter(filters).annotate(
            guestType=Case(
                When(eventstaffmember__event=event, then=Concat(Value('Event Staff: '), 'eventstaffmember__category__name')),
                default_value=Value('Other Staff'),
                output_field=models.CharField()
            )
        ).distinct().values('firstName','lastName','guestType'))

        if self.includeRegistrants and event and self.appliesToEvent(event):
            names += list(Registration.objects.filter(eventregistration__event=event).annotate(
                guestType=Value('Registered',output_field=models.CharField())
            ).values('firstName','lastName','guestType'))
        return names


    def __str__(self):
        return '%s: %s' % (_('Guest list'),self.name)

    class Meta:
        ordering = ('name',)
        verbose_name = _('Guest list')
        verbose_name_plural = _('Guest lists')


class GuestListName(models.Model):
    ''' Additional names to be manually added to a particular guest list '''

    guestList = models.ForeignKey(GuestList,on_delete=models.CASCADE)
    firstName = models.CharField(_('First name'),max_length=50)
    lastName = models.CharField(_('Last name'),max_length=50)

    @property
    def fullName(self):
        return ' '.join([self.firstName or '',self.lastName or ''])
    fullName.fget.short_description = _('Name')

    def __str__(self):
        return '%s: %s' % (_('Guest'), self.fullName)

    class Meta:
        ordering = ('guestList', 'lastName','firstName')
        verbose_name = _('Manually-added guest')
        verbose_name_plural = _('Manually added guests')
        permissions = (
            ('view_guestlist',_('Can view guest lists')),
        )


class GuestListComponent(models.Model):
    guestList = models.ForeignKey(GuestList,on_delete=models.CASCADE)

    staffCategory = models.ForeignKey(EventStaffCategory,verbose_name=_('Category of staff members'),null=True, blank=True, on_delete=models.CASCADE)
    staffMember = models.ForeignKey(StaffMember,verbose_name=_('Individual staff member'),null=True,blank=True,on_delete=models.CASCADE)

    admissionRule = models.CharField(_('Event admission rule'),choices=GUESTLIST_ADMISSION_CHOICES,max_length=10)

    def clean(self):
        ''' Either staffCategory or staffMember must be filled in, but not both. '''
        if not self.staffCategory and not self.staffMember:
            raise ValidationError(_('Either staff category or staff member must be specified.'))
        if self.staffCategory and self.staffMember:
           raise ValidationError(_('Specify either a staff category or a staff member, not both.'))      

    class Meta:
        ordering = ('guestList','admissionRule')
        verbose_name = _('Guest list component')
        verbose_name_plural = _('Guest list components')
        unique_together = ('guestList','staffCategory','staffMember')



# Third Party Imports
from cms.toolbar_base import CMSToolbar
from cms.toolbar.items import Break
from cms.toolbar_pool import toolbar_pool
from django.core.urlresolvers import reverse
from django.utils.translation import ugettext_lazy as _

from .models import GuestList


@toolbar_pool.register
class GuestListToolbar(CMSToolbar):
    ''' Adds submenu for guest lists '''

    def populate(self):
        menu = self.toolbar.get_or_create_menu('core-events', _('Events'))

        break_index = menu.find_first(Break, identifier='related_items_break').index

        if self.request.user.has_perm('guestlist.view_guestlist') or self.request.user.has_perm('guestlist.change_guestlist'):
            guestlist_menu = menu.get_or_create_menu('guestlist', _('Guest Lists'), position=break_index + 1)
        else:
            return

        if self.request.user.has_perm('guestlist.view_guestlist'):
            for thisList in GuestList.objects.all():
                guestlist_menu.add_link_item(thisList.name,
                                        url=reverse('viewGuestList', args=(thisList.id,)))

        if self.request.user.has_perm('guestlist.change_guestlist'):
            guestlist_menu.add_break('manage-break')
            guestlist_menu.add_link_item(_('Manage Guest Lists'),
                                       url=reverse('admin:guestlist_guestlist_changelist'))


from django.contrib import admin
from django.forms import ModelForm, ModelChoiceField, TextInput
from django.utils.translation import ugettext_lazy as _

from dal import autocomplete

from danceschool.core.models import Customer, Registration, TemporaryRegistration
from danceschool.core.admin import CustomerAdmin

from .models import VoucherCategory, Voucher, DanceTypeVoucher, ClassVoucher, SeriesCategoryVoucher, PublicEventCategoryVoucher, SessionVoucher, CustomerGroupVoucher, CustomerVoucher, VoucherUse, TemporaryVoucherUse, VoucherCredit


class CustomerVoucherInlineForm(ModelForm):
    customer = ModelChoiceField(
        queryset=Customer.objects.all(),
        widget=autocomplete.ModelSelect2(
            url='autocompleteCustomer',
            attrs={
                # This will set the input placeholder attribute:
                'data-placeholder': _('Enter a customer name'),
                # This will set the yourlabs.Autocomplete.minimumCharacters
                # options, the naming conversion is handled by jQuery
                'data-minimum-input-length': 2,
                'data-max-results': 4,
                'class': 'modern-style',
            }
        )
    )

    class Meta:
        model = CustomerVoucher
        exclude = []


class CustomerGroupVoucherInline(admin.StackedInline):
    model = CustomerGroupVoucher
    extra = 1
    classes = ['collapse',]


class CustomerVoucherInline(admin.StackedInline):
    model = CustomerVoucher
    form = CustomerVoucherInlineForm
    extra = 1
    classes = ['collapse',]


class CustomerAdminWithVouchers(CustomerAdmin):
    inlines = CustomerAdmin.inlines + [CustomerVoucherInline]


class RegistrationVoucherInline(admin.TabularInline):
    model = VoucherUse
    extra = 0
    readonly_fields = ['voucher','amount']

    # Prevents adding new voucher uses without going through
    # the standard registration process
    def has_add_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False


class TemporaryRegistrationVoucherInline(admin.TabularInline):
    model = TemporaryVoucherUse
    extra = 0
    readonly_fields = ['voucher','amount']

    # Prevents adding new voucher uses without going through
    # the standard registration process
    def has_add_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False


class DanceTypeVoucherInline(admin.StackedInline):
    model = DanceTypeVoucher
    extra = 1
    classes = ['collapse',]


class ClassVoucherInline(admin.StackedInline):
    model = ClassVoucher
    extra = 1
    classes = ['collapse',]


class SeriesCategoryVoucherInline(admin.StackedInline):
    model = SeriesCategoryVoucher
    extra = 1
    classes = ['collapse',]


class PublicEventCategoryVoucherInline(admin.StackedInline):
    model = PublicEventCategoryVoucher
    extra = 1
    classes = ['collapse',]


class EventSessionVoucherInline(admin.StackedInline):
    model = SessionVoucher
    extra = 1
    classes = ['collapse',]


class ClassVoucherInline(admin.StackedInline):
    model = ClassVoucher
    extra = 1
    classes = ['collapse',]


class VoucherUseInline(admin.TabularInline):
    model = VoucherUse
    extra = 0
    readonly_fields = ['registration','amount']

    # Prevents adding new voucher uses without going through
    # the standard registration process.
    def has_add_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False


class VoucherCreditInlineForm(ModelForm):

    class Meta:
        widgets = {
            'description': TextInput,
        }


class VoucherCreditInline(admin.TabularInline):
    model = VoucherCredit
    extra = 1
    fields = ['amount', 'description','creationDate']
    readonly_fields = ['creationDate',]

    form = VoucherCreditInlineForm


class VoucherAdmin(admin.ModelAdmin):
    inlines = [DanceTypeVoucherInline,ClassVoucherInline,SeriesCategoryVoucherInline,PublicEventCategoryVoucherInline,EventSessionVoucherInline,CustomerGroupVoucherInline,CustomerVoucherInline,VoucherUseInline,VoucherCreditInline]
    list_display = ['voucherId','name','category','amountLeft','maxAmountPerUse','expirationDate','isEnabled','restrictions']
    list_filter = ['category','expirationDate','disabled','forFirstTimeCustomersOnly','forPreviousCustomersOnly']
    search_fields = ['voucherId','name','description']
    readonly_fields = ['refundAmount','creationDate']
    actions = ['enableVoucher','disableVoucher']

    fieldsets = (
        (None, {
            'fields': (('voucherId','category'),'name','description',('originalAmount','maxAmountPerUse'),),
        }),
        (_('Voucher Restrictions'), {
            'fields': ('expirationDate',('singleUse','forFirstTimeCustomersOnly','forPreviousCustomersOnly','disabled')),
        }),
        (_('Other Info'), {
            'classes': ('collapse',),
            'fields': ('creationDate','refundAmount'),
        }),
    )

    def isEnabled(self,obj):
        return obj.disabled is False
    isEnabled.short_description = _('Enabled')
    isEnabled.boolean = True

    def restrictions(self,obj):
        text = []
        if obj.singleUse:
            text.append(_('Single use'))
        if obj.forFirstTimeCustomersOnly:
            text.append(_('First-time customer'))
        if obj.forPreviousCustomersOnly:
            text.append(_('Previous customer'))
        if obj.customervoucher_set.all().exists():
            text.append(_('Specific customer'))
        if obj.classvoucher_set.all().exists():
            text.append(_('Specific class'))
        if obj.dancetypevoucher_set.all().exists():
            text.append(_('Specific dance type/level'))
        return ', '.join([str(x) for x in text])
    restrictions.short_description = _('Restrictions')

    def disableVoucher(self, request, queryset):
        rows_updated = queryset.update(disabled=True)
        if rows_updated == 1:
            message_bit = "1 voucher was"
        else:
            message_bit = "%s vouchers were" % rows_updated
        self.message_user(request, "%s successfully disabled." % message_bit)
    disableVoucher.short_description = _('Disable selected Vouchers')

    def enableVoucher(self, request, queryset):
        rows_updated = queryset.update(disabled=False)
        if rows_updated == 1:
            message_bit = "1 voucher was"
        else:
            message_bit = "%s vouchers were" % rows_updated
        self.message_user(request, "%s successfully enabled." % message_bit)
    enableVoucher.short_description = _('Enable selected Vouchers')


# This adds inlines to Registration and TemporaryRegistration without subclassing
admin.site._registry[Registration].inlines.insert(0,RegistrationVoucherInline)
admin.site._registry[TemporaryRegistration].inlines.insert(0,TemporaryRegistrationVoucherInline)

admin.site.register(VoucherCategory)
admin.site.register(Voucher,VoucherAdmin)
admin.site.unregister(Customer)
admin.site.register(Customer,CustomerAdminWithVouchers)
